// SPDX-License-Identifier: GPL-2.0
/*
 * raid6_vx8.c
 *
 * 8-way unrolled RAID6 gen/xor functions for s390
 * based on the vector facility
 *
 * Copyright IBM Corp. 2016
 * Author(s): Martin Schwidefsky <schwidefsky@de.ibm.com>
 *
 * This file is postprocessed using unroll.awk.
 */

#include <linux/raid/pq.h>
#include <asm/fpu/api.h>

asm(".include \"asm/vx-insn.h\"\n");

#define NSIZE 16

static inline void LOAD_CONST(void)
{
	asm volatile("VREPIB %v24,7");
	asm volatile("VREPIB %v25,0x1d");
}

/*
 * The SHLBYTE() operation shifts each of the 16 bytes in
 * vector register y left by 1 bit and stores the result in
 * vector register x.
 */
static inline void SHLBYTE(int x, int y)
{
	asm volatile ("VAB %0,%1,%1" : : "i" (x), "i" (y));
}

/*
 * For each of the 16 bytes in the vector register y the MASK()
 * operation returns 0xFF if the high bit of the byte is 1,
 * or 0x00 if the high bit is 0. The result is stored in vector
 * register x.
 */
static inline void MASK(int x, int y)
{
	asm volatile ("VESRAVB	%0,%1,24" : : "i" (x), "i" (y));
}

static inline void AND(int x, int y, int z)
{
	asm volatile ("VN %0,%1,%2" : : "i" (x), "i" (y), "i" (z));
}

static inline void XOR(int x, int y, int z)
{
	asm volatile ("VX %0,%1,%2" : : "i" (x), "i" (y), "i" (z));
}

static inline void LOAD_DATA(int x, u8 *ptr)
{
	typedef struct { u8 _[16 * 8]; } addrtype;
	register addrtype *__ptr asm("1") = (addrtype *) ptr;

	asm volatile ("VLM %2,%3,0,%1"
		      : : "m" (*__ptr), "a" (__ptr), "i" (x),
			  "i" (x + 8 - 1));
}

static inline void STORE_DATA(int x, u8 *ptr)
{
	typedef struct { u8 _[16 * 8]; } addrtype;
	register addrtype *__ptr asm("1") = (addrtype *) ptr;

	asm volatile ("VSTM %2,%3,0,1"
		      : "=m" (*__ptr) : "a" (__ptr), "i" (x),
			"i" (x + 8 - 1));
}

static inline void COPY_VEC(int x, int y)
{
	asm volatile ("VLR %0,%1" : : "i" (x), "i" (y));
}

static void raid6_s390vx8_gen_syndrome(int disks, size_t bytes, void **ptrs)
{
	struct kernel_fpu vxstate;
	u8 **dptr, *p, *q;
	int d, z, z0;

	kernel_fpu_begin(&vxstate, KERNEL_VXR);
	LOAD_CONST();

	dptr = (u8 **) ptrs;
	z0 = disks - 3;		/* Highest data disk */
	p = dptr[z0 + 1];	/* XOR parity */
	q = dptr[z0 + 2];	/* RS syndrome */

	for (d = 0; d < bytes; d += 8*NSIZE) {
		LOAD_DATA(0,&dptr[z0][d]);
		COPY_VEC(8+0,0+0);
		COPY_VEC(8+1,0+1);
		COPY_VEC(8+2,0+2);
		COPY_VEC(8+3,0+3);
		COPY_VEC(8+4,0+4);
		COPY_VEC(8+5,0+5);
		COPY_VEC(8+6,0+6);
		COPY_VEC(8+7,0+7);
		for (z = z0 - 1; z >= 0; z--) {
			MASK(16+0,8+0);
			MASK(16+1,8+1);
			MASK(16+2,8+2);
			MASK(16+3,8+3);
			MASK(16+4,8+4);
			MASK(16+5,8+5);
			MASK(16+6,8+6);
			MASK(16+7,8+7);
			AND(16+0,16+0,25);
			AND(16+1,16+1,25);
			AND(16+2,16+2,25);
			AND(16+3,16+3,25);
			AND(16+4,16+4,25);
			AND(16+5,16+5,25);
			AND(16+6,16+6,25);
			AND(16+7,16+7,25);
			SHLBYTE(8+0,8+0);
			SHLBYTE(8+1,8+1);
			SHLBYTE(8+2,8+2);
			SHLBYTE(8+3,8+3);
			SHLBYTE(8+4,8+4);
			SHLBYTE(8+5,8+5);
			SHLBYTE(8+6,8+6);
			SHLBYTE(8+7,8+7);
			XOR(8+0,8+0,16+0);
			XOR(8+1,8+1,16+1);
			XOR(8+2,8+2,16+2);
			XOR(8+3,8+3,16+3);
			XOR(8+4,8+4,16+4);
			XOR(8+5,8+5,16+5);
			XOR(8+6,8+6,16+6);
			XOR(8+7,8+7,16+7);
			LOAD_DATA(16,&dptr[z][d]);
			XOR(0+0,0+0,16+0);
			XOR(0+1,0+1,16+1);
			XOR(0+2,0+2,16+2);
			XOR(0+3,0+3,16+3);
			XOR(0+4,0+4,16+4);
			XOR(0+5,0+5,16+5);
			XOR(0+6,0+6,16+6);
			XOR(0+7,0+7,16+7);
			XOR(8+0,8+0,16+0);
			XOR(8+1,8+1,16+1);
			XOR(8+2,8+2,16+2);
			XOR(8+3,8+3,16+3);
			XOR(8+4,8+4,16+4);
			XOR(8+5,8+5,16+5);
			XOR(8+6,8+6,16+6);
			XOR(8+7,8+7,16+7);
		}
		STORE_DATA(0,&p[d]);
		STORE_DATA(8,&q[d]);
	}
	kernel_fpu_end(&vxstate, KERNEL_VXR);
}

static void raid6_s390vx8_xor_syndrome(int disks, int start, int stop,
					size_t bytes, void **ptrs)
{
	struct kernel_fpu vxstate;
	u8 **dptr, *p, *q;
	int d, z, z0;

	dptr = (u8 **) ptrs;
	z0 = stop;		/* P/Q right side optimization */
	p = dptr[disks - 2];	/* XOR parity */
	q = dptr[disks - 1];	/* RS syndrome */

	kernel_fpu_begin(&vxstate, KERNEL_VXR);
	LOAD_CONST();

	for (d = 0; d < bytes; d += 8*NSIZE) {
		/* P/Q data pages */
		LOAD_DATA(0,&dptr[z0][d]);
		COPY_VEC(8+0,0+0);
		COPY_VEC(8+1,0+1);
		COPY_VEC(8+2,0+2);
		COPY_VEC(8+3,0+3);
		COPY_VEC(8+4,0+4);
		COPY_VEC(8+5,0+5);
		COPY_VEC(8+6,0+6);
		COPY_VEC(8+7,0+7);
		for (z = z0 - 1; z >= start; z--) {
			MASK(16+0,8+0);
			MASK(16+1,8+1);
			MASK(16+2,8+2);
			MASK(16+3,8+3);
			MASK(16+4,8+4);
			MASK(16+5,8+5);
			MASK(16+6,8+6);
			MASK(16+7,8+7);
			AND(16+0,16+0,25);
			AND(16+1,16+1,25);
			AND(16+2,16+2,25);
			AND(16+3,16+3,25);
			AND(16+4,16+4,25);
			AND(16+5,16+5,25);
			AND(16+6,16+6,25);
			AND(16+7,16+7,25);
			SHLBYTE(8+0,8+0);
			SHLBYTE(8+1,8+1);
			SHLBYTE(8+2,8+2);
			SHLBYTE(8+3,8+3);
			SHLBYTE(8+4,8+4);
			SHLBYTE(8+5,8+5);
			SHLBYTE(8+6,8+6);
			SHLBYTE(8+7,8+7);
			XOR(8+0,8+0,16+0);
			XOR(8+1,8+1,16+1);
			XOR(8+2,8+2,16+2);
			XOR(8+3,8+3,16+3);
			XOR(8+4,8+4,16+4);
			XOR(8+5,8+5,16+5);
			XOR(8+6,8+6,16+6);
			XOR(8+7,8+7,16+7);
			LOAD_DATA(16,&dptr[z][d]);
			XOR(0+0,0+0,16+0);
			XOR(0+1,0+1,16+1);
			XOR(0+2,0+2,16+2);
			XOR(0+3,0+3,16+3);
			XOR(0+4,0+4,16+4);
			XOR(0+5,0+5,16+5);
			XOR(0+6,0+6,16+6);
			XOR(0+7,0+7,16+7);
			XOR(8+0,8+0,16+0);
			XOR(8+1,8+1,16+1);
			XOR(8+2,8+2,16+2);
			XOR(8+3,8+3,16+3);
			XOR(8+4,8+4,16+4);
			XOR(8+5,8+5,16+5);
			XOR(8+6,8+6,16+6);
			XOR(8+7,8+7,16+7);
		}
		/* P/Q left side optimization */
		for (z = start - 1; z >= 0; z--) {
			MASK(16+0,8+0);
			MASK(16+1,8+1);
			MASK(16+2,8+2);
			MASK(16+3,8+3);
			MASK(16+4,8+4);
			MASK(16+5,8+5);
			MASK(16+6,8+6);
			MASK(16+7,8+7);
			AND(16+0,16+0,25);
			AND(16+1,16+1,25);
			AND(16+2,16+2,25);
			AND(16+3,16+3,25);
			AND(16+4,16+4,25);
			AND(16+5,16+5,25);
			AND(16+6,16+6,25);
			AND(16+7,16+7,25);
			SHLBYTE(8+0,8+0);
			SHLBYTE(8+1,8+1);
			SHLBYTE(8+2,8+2);
			SHLBYTE(8+3,8+3);
			SHLBYTE(8+4,8+4);
			SHLBYTE(8+5,8+5);
			SHLBYTE(8+6,8+6);
			SHLBYTE(8+7,8+7);
			XOR(8+0,8+0,16+0);
			XOR(8+1,8+1,16+1);
			XOR(8+2,8+2,16+2);
			XOR(8+3,8+3,16+3);
			XOR(8+4,8+4,16+4);
			XOR(8+5,8+5,16+5);
			XOR(8+6,8+6,16+6);
			XOR(8+7,8+7,16+7);
		}
		LOAD_DATA(16,&p[d]);
		XOR(16+0,16+0,0+0);
		XOR(16+1,16+1,0+1);
		XOR(16+2,16+2,0+2);
		XOR(16+3,16+3,0+3);
		XOR(16+4,16+4,0+4);
		XOR(16+5,16+5,0+5);
		XOR(16+6,16+6,0+6);
		XOR(16+7,16+7,0+7);
		STORE_DATA(16,&p[d]);
		LOAD_DATA(16,&q[d]);
		XOR(16+0,16+0,8+0);
		XOR(16+1,16+1,8+1);
		XOR(16+2,16+2,8+2);
		XOR(16+3,16+3,8+3);
		XOR(16+4,16+4,8+4);
		XOR(16+5,16+5,8+5);
		XOR(16+6,16+6,8+6);
		XOR(16+7,16+7,8+7);
		STORE_DATA(16,&q[d]);
	}
	kernel_fpu_end(&vxstate, KERNEL_VXR);
}

static int raid6_s390vx8_valid(void)
{
	return MACHINE_HAS_VX;
}

const struct raid6_calls raid6_s390vx8 = {
	raid6_s390vx8_gen_syndrome,
	raid6_s390vx8_xor_syndrome,
	raid6_s390vx8_valid,
	"vx128x8",
	1
};
